// ledarray_slave.c
// for NerdKits with ATmega168
// hevans@nerdkits.com

#define F_CPU 14745600

#include <stdio.h>

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <inttypes.h>

#include "../libnerdkits/delay.h"
// PIN DEFINITIONS: 
//
// PB1 - ROW 0
// PC0-3 - ROW 1-4
// PC4-6, PD0-7 COL 0-9
#define ROWS 5
#define COLS 20

volatile uint8_t la_row, real_row;
volatile uint8_t la_data_array[COLS];
volatile uint8_t la_data_buffer_array[COLS];

//pointers for the buffers - to allow us to flip easily
volatile uint8_t* la_data;
volatile uint8_t* la_data_buffer;

//flip the buffers
inline void flip(){
  uint8_t* temp;
  temp = la_data;
  la_data = la_data_buffer;
  la_data_buffer = temp;
}


inline uint8_t ledarray_get(uint8_t i, uint8_t j) {
  if(i < ROWS && j < COLS) {
    if((la_data[j] & (1<<i)) != 0) {
      return 1;
    } else {
      return 0;
    }
  } else {
    return 0;
  }
}

inline void ledarray_set(uint8_t i, uint8_t j, uint8_t onoff) {
  if(i < ROWS && j < COLS) {
    if(onoff) {
      la_data[j] |= (1<<i);
    } else {
      la_data[j] &= ~(1<<i);
    }
  }
}

inline void ledarray_set_columndriver(uint8_t j, uint8_t onoff, uint8_t sense) {
  // cols 0-1: PC4-5
  // cols 2-9: PD0-7
  if(j < 2) {
    if(onoff) {
      PORTC |= (1<<(PC4 + j));
    } else {
      PORTC &= ~(1<<(PC4 + j));
    }
    if(sense == onoff) {
      DDRC |= (1<<(PC4 + j));
    } else {
      DDRC &= ~(1<<(PC4 + j));
      PORTC &= ~(1<<(PC4 + j));
    }
  } else {
    if(onoff) {
      PORTD |= (1<<(PD0+(j-2)));
    } else {
      PORTD &= ~(1<<(PD0+(j-2)));
    }  
    if(sense == onoff) {
      DDRD |= (1<<(PD0 + (j-2)));
    } else {
      DDRD &= ~(1<<(PD0+(j-2)));
      PORTD &= ~(1<<(PD0 +(j-2)));
    }
  }
}

inline void ledarray_all_off() {
  // turn off all row drivers
  DDRB &= ~(1<<PB1);
  PORTB &= ~(1<<PB1);
    
  // turn off all column drivers
  DDRC &= ~( (1<<PC0) | (1<<PC1) | (1<<PC2) | (1<<PC3) | (1<<PC4) | (1<<PC5) );
  PORTC &= ~( (1<<PC0) | (1<<PC1) | (1<<PC2) | (1<<PC3) | (1<<PC4) | (1<<PC5) );
  DDRD &= ~( (1<<PD0) | (1<<PD1) | (1<<PD2) | (1<<PD3) | (1<<PD4) | (1<<PD5) | (1<<PD6) | (1<<PD7) );
  PORTD &= ~( (1<<PD0) | (1<<PD1) | (1<<PD2) | (1<<PD3) | (1<<PD4) | (1<<PD5) | (1<<PD6) | (1<<PD7) );  
}

inline void set_row_driver(uint8_t real_row,uint8_t sink){
  uint8_t temp_row_select = 0;
  if(real_row == 0){
    temp_row_select = PB1;
    if(sink){
      // activate row driver SINK
      PORTB &= ~(1 << temp_row_select);
      DDRB |= (1 << temp_row_select);
    } else {
      // activate row driver SOURCE
      PORTB |= (1 << temp_row_select);
      DDRB |= (1 << temp_row_select);
    }
  } else {
    temp_row_select = PC0 + (real_row - 1);
    if(sink){
      // activate row driver SINK
      PORTC &= ~(1 << temp_row_select);
      DDRC |= (1 << temp_row_select);
    } else {
      // activate row driver SOURCE
      PORTC |= (1 << temp_row_select);
      DDRC |= (1 << temp_row_select);
    }
  } 
}

ISR(TIMER0_OVF_vect) {
  sei();
  // turn off old row driver
  //DDRB &= ~(1 << (PB1 + real_row));
  //PORTB &= ~(1 << (PB1 + real_row));
  ledarray_all_off();

  // increment row number
  if(++la_row == 2*ROWS)
    la_row = 0;

  // set column drivers appropriately
  uint8_t j;
  if(la_row%2 == 0) {
    // even la_row number: fill even columns
    real_row = la_row / 2;

    for(j=0; j<COLS/2; j++) {
      ledarray_set_columndriver(j, ledarray_get(real_row, 2*j), 1);
    }
    //activate the row driver, sink
    set_row_driver(real_row,1);
  } else {
    // odd la_row number: fill odd columns
    real_row = (la_row-1)/2;
    for(j=0; j<COLS/2; j++) {
      ledarray_set_columndriver(j, 1 - ledarray_get(real_row, 2*j + 1), 0);
    }
    //activate the row driver, source
    set_row_driver(real_row,0);
  }  
}

void ledarray_init() {

  //set up the pointer for the buffers
  la_data = &la_data_array[0];
  la_data_buffer = &la_data_buffer_array[0];

  // Timer0 CK/64 (900Hz)
  TCCR0B = (1<<CS01) | (1<<CS00);
  TIMSK0 = (1<<TOIE0);
  
  // outputs (set row drivers high for off)
  DDRC &= ~( (1<<PC0) | (1<<PC1) | (1<<PC2) | (1<<PC3) | (1<<PC4) | (1<<PC5) );
  DDRD &= ~( (1<<PD0) | (1<<PD1) | (1<<PD2) | (1<<PD3) | (1<<PD4) | (1<<PD5) | (1<<PD6) | (1<<PD7) );
  DDRB &= ~(1<<PB1);
}

void ledarray_left_shift() {
  // shift everything one position left
  uint8_t i, j;
  for(i=0; i<ROWS; i++) {
    for(j=0; j<COLS-1; j++) {
      ledarray_set(i,j, ledarray_get(i, j+1));
    }
  }
}

void ledarray_blank() {
  uint8_t i, j;
  for(i=0; i<ROWS; i++) {
    for(j=0; j<COLS; j++) {
      ledarray_set(i,j,0);
    }
  }
}

uint8_t c;
uint8_t buff_pos;
ISR(SPI_STC_vect){
  //PORTC |= (1<<PC4);
  c = SPDR;
  if(c == 0xff){
    buff_pos = 0;
    flip();
    return;
  } else if(buff_pos > COLS-1){
    return;
  } else {
    la_data_buffer[buff_pos] = c;
    buff_pos++;
  }
}

void slave_init(){
  //debug LED
  //DDRC |= (1<<PC4);

  //set MISO
  DDRB |= (1<<PB4);
  SPCR |= (1<<SPE) | (1<<SPIE);
}

int main() {

  //disable the uart module
  UCSR0B = 0;  

  ledarray_init();

  slave_init();
  buff_pos = 0; 
  // activate interrupts
  sei();

  uint8_t i,j;
  while(1) {
    delay_ms(10);
//    for(i=0;i<ROWS;i++){
//      for(j=0;j<COLS;j++){
//        ledarray_set(i,j,1);
//        delay_ms(100);
//      }
//    }
//    for(i=0;i<ROWS;i++){
//      for(j=0;j<COLS;j++){
//        ledarray_set(i,j,0);
//        delay_ms(100);
//      }
//    }
  }
  return 0;
}

